"""
Module containing a dynamic growth and flux estimation model
"""
from __future__ import annotations

import numpy as np
from scipy.integrate import solve_ivp

from physiofit.models.base_model import Model, Bounds

class ChildModel(Model):

    def __init__(self, data):

        super().__init__(data)
        self.model_name = "Dynamic Monod model (1 substrate, 1 product)"
        self.vini = 1
        self.parameters_to_estimate = None

    def get_params(self):

        self.parameters_to_estimate = {
            "X_0": self.vini,
            "y_BM": self.vini
        }

        self.bounds = Bounds(
            X_0=(1e-3, 10),
            y_BM=(1e-3, 3)
        )

        for metabolite in self.metabolites:
            if metabolite.startswith("S_"):
                self.parameters_to_estimate.update(
                    {
                        f"{metabolite}_km": self.vini,
                        f"{metabolite}_qsmax": self.vini,
                        f"{metabolite}_s_0": 100
                    }
                )
                self.bounds.update(
                    {
                        f"{metabolite}_km": (1e-6, 100),
                        f"{metabolite}_qsmax": (1e-6, 100),
                        f"{metabolite}_s_0": (1e-6, 100)
                    }
                )
                break
                
        for metabolite in self.metabolites:
            if metabolite.startswith("P_"):
                self.parameters_to_estimate.update(
                    {
                        f"{metabolite}_y_P": self.vini,
                        f"{metabolite}_p_0": 100
                    }

                )
                self.bounds.update(
                    {
                        f"{metabolite}_y_P": (1e-6, 100),
                        f"{metabolite}_p_0": (1e-6, 100)
                    }
                )
                break

        if len(self.parameters_to_estimate) != 7:
            raise ValueError(
                "This model expects 2 metabolites in the data file (1 substrate with name starting with 'S_' and 1 "
                "product with name starting with 'P_')."
            )


    @staticmethod
    def simulate(
            params_opti: list,
            data_matrix: np.ndarray,
            time_vector: np.ndarray,
            params_non_opti: dict
    ):

        # Get parameters
        x_0 = params_opti[0]
        y_BM = params_opti[1]
        km = params_opti[2]
        qsmax = params_opti[3]
        s_0 = params_opti[4]
        y_P = params_opti[5]
        p_0 = params_opti[6]
        params = (y_BM, y_P, km, qsmax)

        # initialize variables at t=0
        state = [x_0, s_0, p_0]
        
        def calculate_derivative(t, state, y_BM, y_P, km, qsmax):
            
            # get substrate and biomass concentrations
            s_t = state[0]
            x_t = state[1]

            # calculate fluxes
            qs_t = qsmax * (s_t / (km + s_t))
            mu_t = y_BM * qs_t
            qp_t = y_P * qs_t

            # calculate derivatives
            dx = mu_t * x_t
            ds = -qs_t * x_t
            dp = qp_t * x_t

            return dx, ds, dp

        # simulate time-course concentrations
        sol = solve_ivp(
            fun=calculate_derivative,
            t_span=(np.min(time_vector), np.max(time_vector)),
            y0 = state,
            args=params,
            method="LSODA",
            t_eval = list(time_vector)
        )

        return sol.y.T

if __name__ == "__main__":

    from physiofit.base.io import IoHandler

    io = IoHandler()
    data = io.read_data(r"C:\Users\legregam\PycharmProjects\PhysioFit\data\KEIO_test_data\ode_test\KEIO_ROBOT6_1.tsv")
    data = data.sort_values("time")

    model = ChildModel(data)
    model.get_params()
    params = [param for param in model.parameters_to_estimate.values()]
    sol = ChildModel.simulate(
        params,
        model.data,
        model.time_vector,
        None
    )
    print(f"Times: {sol.t}")
    print(f"Values: {np.array(sol.y)}")



"""
Test the creation and use of the PhysioFitter
"""
import numpy as np
import pandas as pd
import pytest
import physiofit


def test_physiofitter(data, sds):
    """
    Test that the model and PhysioFitter can be safely instanciated from IoHandler
    """

    io = physiofit.base.io.IoHandler()
    model = io.select_model("Steady-state batch model", data)
    assert isinstance(model, physiofit.models.base_model.Model)
    model.get_params()
    fitter = io.initialize_fitter(
        model.data,
        model=model,
        sd=sds,
        debug_mode=True
    )
    assert isinstance(fitter, physiofit.base.fitter.PhysioFitter)

def test_simulation(data, sds):

    io = physiofit.base.io.IoHandler()
    model = io.select_model("Steady-state batch model", data)
    model.get_params()
    sim_mat = model.simulate(
        [param for param in model.parameters_to_estimate.values()],
        model.experimental_matrix,
        model.time_vector,
        model.fixed_parameters
    )
    assert isinstance(sim_mat, np.ndarray)

def test_wrong_entry_for_model_data(data, sds):

    with pytest.raises(AttributeError):
        io = physiofit.base.io.IoHandler()
        model = io.select_model("Steady-state batch model", None)
    with pytest.raises(AttributeError):
        model = io.select_model(
            "Steady-state batch model",
            np.array(
                [[0, 1, 2],[0, 1, 2]]
            )
        )
    with pytest.raises(AttributeError):
        model = io.select_model(
            "Steady-state batch model",
            "Hello world this is an error"
        )

def test_optimization_process(data, sds):

    io = physiofit.base.io.IoHandler()
    model = io.select_model("Steady-state batch model", data)
    model.get_params()
    fitter = io.initialize_fitter(
        model.data,
        model=model,
        sd=sds,
        debug_mode=True
    )
    fitter.optimize()
    assert isinstance(fitter.simulated_data, pd.DataFrame)
    for col in ["X", "Glucose", "Acetate"]:
        assert col in fitter.simulated_data.columns

def test_monte_carlo(data, sds):
    io = physiofit.base.io.IoHandler()
    model = io.select_model("Steady-state batch model", data)
    model.get_params()
    fitter = io.initialize_fitter(
        model.data,
        model=model,
        sd=sds,
        debug_mode=True
    )
    fitter.optimize()
    fitter.monte_carlo_analysis()
    assert hasattr(fitter, "matrices_ci")
    assert isinstance(fitter.matrices_ci["lower_ci"], np.ndarray)
    assert isinstance(fitter.matrices_ci["higher_ci"], np.ndarray)
    assert np.subtract(fitter.matrices_ci["higher_ci"], fitter.matrices_ci["lower_ci"]).all() > 0
    assert hasattr(fitter, "parameter_stats")
    assert isinstance(fitter.parameter_stats, dict)

def test_that_simulated_and_experimental_matrices_are_close(data, sds):
    io = physiofit.base.io.IoHandler()
    model = io.select_model("Steady-state batch model", data)
    model.get_params()
    fitter = io.initialize_fitter(
        model.data,
        model=model,
        sd=sds,
        debug_mode=True
    )
    fitter.optimize()
    assert np.allclose(
        a=fitter.experimental_matrix,
        b=fitter.simulated_data,
        rtol=1,
        equal_nan=True
    )




"""
Module containing the methods used by PhysioFit.
"""
from __future__ import annotations

import numpy as np
import pandas as pd

from physiofit.models.base_model import Model, Bounds

class ChildModel(Model):

    def __init__(self, data):

        super().__init__(data)
        self.model_name = "Steady-state batch model with lag phase and " \
                          "degradation of metabolites "
        self.vini = 1
        self.parameters_to_estimate = None
        self.fixed_parameters = None

    def get_params(self):

        self.parameters_to_estimate = {
            "X_0" : self.vini,
            "growth_rate" : self.vini,
            "t_lag" : self.vini
        }
        self.bounds = Bounds(
            X_0=(1e-3, 10),
            growth_rate=(1e-3, 3),
            t_lag = (0, 0.5*self.time_vector.max())
        )
        for metabolite in self.metabolites:
            self.parameters_to_estimate.update(
                {
                    f"{metabolite}_q" : self.vini,
                    f"{metabolite}_M0" : self.vini
                }
            )
            self.bounds.update(
                {
                    f"{metabolite}_q": (-50, 50),
                    f"{metabolite}_M0": (1e-6, 50)
                }
            )

        self.fixed_parameters = {"Degradation": {
            met: 0 for met in self.metabolites
            }
        }

    @staticmethod
    def simulate(
            params_opti: list,
            data_matrix: np.ndarray,
            time_vector: np.ndarray,
            params_non_opti: dict
    ):
        # Get end shape
        simulated_matrix = np.empty_like(data_matrix)

        # Get initial params
        x_0 = params_opti[0]
        mu = params_opti[1]
        t_lag = params_opti[2]

        # We get indices in time vector where time < t_lag
        idx = np.nonzero(time_vector < t_lag)

        # Fill at those indices with x_0
        x_t_lag = np.full((len(idx) - 1,), x_0)

        # The rest of the biomass points are calculated as usual
        mult_by_time = x_0 * np.exp(mu * (time_vector[len(idx) - 1:] - t_lag))
        simulated_matrix[:, 0] = np.concatenate((x_t_lag, mult_by_time),
                                                axis=None)
        fixed_params = [value for value in params_non_opti["Degradation"].values()]

        for i in range(1, int(len(params_opti) / 2)):
            q = params_opti[i * 2 + 1]
            m_0 = params_opti[i * 2 + 2]
            k = fixed_params[i - 1]
            m_t_lag = np.full((len(idx) - 1,), m_0)
            mult_by_time = q * (x_0 / (mu + k)) * (np.exp(
                mu * (time_vector - t_lag)
            ) - np.exp(-k * (time_vector - t_lag))) + (m_0 * np.exp(
                -k * time_vector)
                                                       )
            simulated_matrix[:, i] = np.concatenate(
                (m_t_lag, mult_by_time),
                axis=None
            )

        return simulated_matrix


if __name__ == "__main__":
    model = ChildModel(
        pd.read_csv(
            r"C:\Users\legregam\Documents\Projets\PhysioFit\data"
            r"\KEIO_test_data\KEIO_ROBOT6_1\KEIO_ROBOT6_1.tsv",
            sep='\t')
    )
    print(type(model))



"""
Module containing the Command-Line Interface control logic.

The Command-Line Interface has two goals:

    i) Have a local CLI that can be used to launch jobs in a concise and quick manner, without use of the Graphical
    User Interface. This can help when testing many different parameter sets, and also permits easy pipelining of the
    tool, as described below.
    ii) Have a version compatible with an integration onto the Galaxy W4M platform to be used in automated flux
    calculation workflows

"""
import argparse
import logging
from pathlib import Path
import sys
# import shutil

import pandas as pd

from physiofit.base.io import IoHandler, StandardDevs, ConfigParser

def args_parse():
    """
    Parse arguments from user input.
    :return: Argument Parser object
    :rtype: class: argparse.ArgumentParser
    """

    parser = argparse.ArgumentParser(
        "Physiofit: Extracellular flux estimation software"
    )

    # Parse data arguments (tsv + json)
    parser.add_argument(
        "-d", "--data", type=str,
        help="Path to data file in tabulated format (txt or tsv)"
    )
    parser.add_argument(
        "-c", "--config", type=str,
        help="Path to config file in yaml format"
    )
    parser.add_argument(
        "-m", "--model", type=str,
        help="Which model should be chosen. Useful only if generating related config file"
    )

    # Parse developer arguments
    parser.add_argument(
        "-g", "--galaxy", action="store_true",
        help="Is the CLI being used on the galaxy platform"
    )
    parser.add_argument(
        "--list", action="store_true",
        help="Return the list of models in model folder"
    )
    parser.add_argument(
        "-v", "--debug_mode", action="store_true",
        help="Activate the debug logs"
    )

    # Parse selective output path arguments (for galaxy implementation mostly)
    parser.add_argument(
        "-op", "--output_pdf", type=str,
        help="Path to output the pdf file containing plots"
    )
    parser.add_argument(
        "-of", "--output_fluxes", type=str,
        help="Path to output the flux results"
    )
    parser.add_argument(
        "-os", "--output_stats", type=str,
        help="Path to output the khi² test"
    )
    parser.add_argument(
        "-oc", "--output_config", type=str,
        help="Path to output the yaml config file"
    )
    parser.add_argument(
        "-or", "--output_recap", type=str,
        help="Path to output the summary"
    )
    parser.add_argument(
        "-oz", "--output_zip", type=str,
        help="Path to export zip file"
    )

    return parser

def run(data, args, logger, experiments):

    for exp in experiments:
        io = IoHandler()
        exp_data = data.loc[exp, :].sort_values("time").copy()
        logger.info(f"Input Data: \n{exp_data}")
        io.home_path = Path(args.data).parent
        logger.info(f"Home path: {io.home_path}")
        logger.info(f"Reading configuration file at {args.config}")
        io.configparser = io.read_yaml(args.config)
        logger.info(f"Config File:\n{io.configparser}")
        model = io.select_model(io.configparser.model["model_name"], exp_data)
        model.get_params()
        logger.info(f"Selected Model:\n{model}")
        model = io.configparser.update_model(model)
        logger.info(f"Updated Model: \n{model}")
        logger.info(f"Model Data: \n{model.data}")

        fitter = io.initialize_fitter(
            model.data,
            model=model,
            mc=io.configparser.mc,
            iterations=io.configparser.iterations,
            sd=io.configparser.sds,
            debug_mode=args.debug_mode
        )
        fitter.optimize()
        if fitter.mc:
            fitter.monte_carlo_analysis()
        fitter.khi2_test()
        df = pd.DataFrame.from_dict(
            fitter.parameter_stats,
            orient="columns"
        )
        df.index = [
            f"{exp} {param}" for param in fitter.model.parameters_to_estimate.keys()
        ]
        if not io.multiple_experiments:
            io.multiple_experiments = []
        io.multiple_experiments.append(df)
        if exp is not None:
            res_path = io.home_path / (io.home_path.name + "_res") / exp
        else:
            res_path = io.home_path / (io.home_path.name + "_res")
        logger.info(res_path)
        if not res_path.is_dir():
            res_path.mkdir(parents=True)
        logger.info(f"Results:\n{df}")
        io.output_report(fitter, str(res_path))
        io.output_plots(fitter, str(res_path))
        io.output_pdf(fitter, str(res_path))
        io.figures = []
    if args.output_zip:
        # logger.info(io.home_path)
        # logger.info(io.home_path / io.home_path.name / "_res")
        # logger.info(args.output_zip)
        dir = res_path.parents[0]
        # for line in dir.rglob("*"):
        #     print(line)
        generate_zips(str(dir), args.output_zip, logger)

    # shutil.make_archive(
    #     args.output_zip,
    #     format='zip',
    #     root_dir=str(io.home_path / (io.home_path.name + "_res"))
    #     )
    io.output_recap(export_path=args.output_recap, galaxy=True)

def generate_zips(path_to_data_folder, output_path, logger):
    """Generate output zip file containing results

    Args:
        path_to_data_folder (str): path to folder containing directories & files to zip
        output_path (str): path to export archive to
        logger (Logging.logger): main logger

    Returns:
        None:
    """
    from zipfile import ZipFile
    directory = Path(path_to_data_folder)
    output = Path(output_path)
    with ZipFile(str(output), mode="w") as archive:
        for file_path in directory.rglob("*"):
            archive.write(
                file_path,
                arcname=file_path.relative_to(directory)
        )
    return 0

def generate_config(args, data, logger):

    logger.info("Launching in configuration file generator mode")
    # Run checks
    if args.model is None:
        # Must be path to model.py file or name of a model
        raise ValueError(
            f"Please select a model to generate the associated configuration file"
        )
    if args.output_config is None:
        raise ValueError(
            "Please select an output directory for the configuration file"
        )
    try:
        io = IoHandler()
        if Path(args.model).is_file():
            model_class = io.read_model(args.model)
            model = model_class(data)
            model.get_params()

        else:
            model = io.select_model(args.model, data)
            model.get_params()
    except Exception:
        logger.error("There was an error while initialising the model")
        raise

    # Build config parser and export configuration file
    default_sds = StandardDevs({col: 0.2 for col in model.name_vector})
    configparser = ConfigParser(
        path_to_data=args.data,
        selected_model=model,
        sds=default_sds,
        mc=True,
        iterations=100
    )
    configparser.export_config(args.output_config)
    logger.info(f"Finished exporting the configuration file to {args.output_config}")
    sys.exit()

def process(args):

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    out_stream = logging.StreamHandler()
    formatter = logging.Formatter()
    out_stream.setFormatter(formatter)
    if args.debug_mode:
        out_stream.setLevel(logging.DEBUG)
    else:
        out_stream.setLevel(logging.INFO)
    logger.addHandler(out_stream)

    if args.list:
        IoHandler.get_model_list()
        sys.exit()

    if args.data is None:
        raise ValueError("A path towards the data must be given")

    # Ensure that the path to data is correct
    path_to_data = Path(args.data)
    if not path_to_data.is_file():
        raise ValueError(
            f"The data path is not correct. Please check and try again. Input data path: \n{args.data}"
        )
    # Ensure that the input file is a tsv if we are local
    if not args.galaxy:
        if not path_to_data.suffix in [".tsv", ".txt"]:
            raise TypeError(
                f"The input data must be in tsv/txt format. Detected format: {path_to_data.suffix}"
            )
    else:
        if not path_to_data.suffix == ".dat":
            raise TypeError(
                f"The input data must be in dat format (galaxy data format). Detected format: {path_to_data.suffix}"
            )

    # Read & check data
    data = IoHandler.read_data(str(path_to_data))

    # If no configuration file is present we assume the user wants to generate a default one
    if args.config is None:
        generate_config(args, data, logger)

    # If configuration file is present we launch the optimization
    experiments = list(data["experiments"].unique())
    data = data.set_index("experiments")
    for exp in experiments:
        logger.info(f"Running optimization for {exp}")
        run_data = data.loc[exp, :].sort_values("time").copy()
        run(run_data, args, logger, exp)
    run(data, args, logger, experiments)
    logger.info("Done!")
    sys.exit()

def main():
    parser = args_parse()
    args = parser.parse_args()
    process(args)



"""
PhysioFit software main module
"""

import logging

import numpy as np
from pandas import DataFrame
from scipy.optimize import minimize, differential_evolution
from scipy.stats import chi2

from physiofit.models.base_model import Model

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# TODO: add estimate deg function (eq 6) with plot of best fit and measured values


class PhysioFitter:
    """
    This class is responsible for most of Physiofit's heavy lifting. Features included are:

        * loading of data from **csv** or **tsv** file
        * **equation system initialization** using the following analytical functions (in absence of lag and
          degradation:

            X(t) = X0 * exp(mu * t)
            Mi(t) = qMi * (X0 / mu) * (exp(mu * t) - 1) + Mi0

        * **simulation of data points** from given initial parameters
        * **cost calculation** using the equation:

            residuum = sum((sim - meas) / sd)²

        * **optimization of the initial parameters** using `scipy.optimize.minimize ('Differential evolution', with polish with 'L-BFGS-B' method) <https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html#optimize-minimize-lbfgsb>`_
        * **sensitivity analysis, khi2 tests and plotting**

    :param data: DataFrame containing data and passed by IOstream object
    :type data: class: pandas.DataFrame
    :param model: Model to initialize parameters and optimize
    :type model: physiofit.models.base_model.Model
    :param mc: Should Monte-Carlo sensitivity analysis be performed (default=True)
    :type mc: Boolean
    :param iterations: number of iterations for Monte-Carlo simulation (default=50)
    :type iterations: int
    :param sd: sd matrix used for residuum calculations. Can be:

                * a matrix with the same dimensions as the measurements matrix (but without the time column)
                * a named vector containing sds for all the metabolites provided in the input file
                * 0  in which case the matrix is automatically constructed from default values
                * a dictionary with the data column headers as keys and the associated value as a scalar or list

    :type sd: int, float, list, dict or ndarray
    """

    def __init__(
            self,
            data,
            model,
            mc=True,
            iterations=100,
            sd=None,
            debug_mode=False
    ):

        self.data = data
        self.model = model
        self.mc = mc
        self.iterations = iterations
        self.sd = sd
        self.debug_mode = debug_mode
        self.experimental_matrix = self.data.drop("time", axis=1).to_numpy()

        self.simulated_matrix = None
        self.simulated_data = None
        self.optimize_results = None
        self.simulate = None
        self.parameter_stats = None
        self.opt_params_sds = None
        self.matrices_ci = None
        self.opt_conf_ints = None
        self.khi2_res = None

    def verify_attrs(self):
        """Check that attributes are valid"""

        if type(self.iterations) is not int:
            raise TypeError(
                f"Number of monte carlo iterations must be an "
                f"integer, and not of type {type(self.iterations)}"
            )

        allowed_sds = [int, float, list, np.ndarray]
        if type(self.sd) not in allowed_sds:
            raise TypeError(
                f"sds is not in the right format ({type(self.sd)}. "
                f"Compatible formats are:\n{allowed_sds}"
            )

    def _sd_dict_to_matrix(self):
        """Convert sd dictionary to matrix/vector"""

        logger.debug("Initializing sd matrix from dict")
        # Perform checks
        for key in self.sd.keys():
            if key not in self.model.name_vector:
                raise KeyError(
                    f"The key {key} is not part of the data headers"
                )
        for name in self.model.name_vector:
            if name not in self.sd.keys():
                raise KeyError(
                    f"The key {name} is missing from the sds dict"
                )

        # Get lengths of each sd entry
        sd_lengths = [
            len(self.sd[key]) if type(self.sd[key]) not in [float, int] else 1
            for key in self.sd.keys()
        ]

        # Make sure that lengths are the same
        if not all(elem == sd_lengths[0] for elem in sd_lengths):
            raise ValueError("All sd vectors must have the same length")

        # Build matrix/vector
        if sd_lengths[0] == 1:
            self.sd = [self.sd[name] for name in self.model.name_vector]
        else:
            columns = (self.sd[name] for name in self.model.name_vector)
            matrix = np.column_stack(columns)
            self.sd = matrix

        logger.debug(f"SD object type: {type(self.sd)}")
        logger.debug(f"Array built from dict:\n{self.sd}")

    def initialize_sd_matrix(self):
        """
        Initialize the sd matrix from different types of inputs: single value,
        vector or matrix.

        :return: None
        """

        # This function can be optimized, if the input is a matrix we should
        # detect it directly
        logger.info("Initializing sd matrix...\n")

        # If sd is None, we generate the default matrix
        if self.sd is None or self.sd == {}:
            try:
                self.sd = {"X": 0.2}
                for col in self.data.columns[2:]:
                    self.sd.update({col: 0.5})
            except Exception:
                raise

        if isinstance(self.sd, dict):
            self._sd_dict_to_matrix()
        # When sd is a single value, we build a sd matrix containing the value
        # in all positions
        if isinstance(self.sd, int) or isinstance(self.sd, float):
            self._build_sd_matrix()
            logger.debug(f"SD matrix: {self.sd}\n")
            return
        if not isinstance(self.sd, np.ndarray) and not isinstance(self.sd,
                                                                  list):
            raise TypeError(
                f"Cannot coerce SD to array. Please check that a list or array "
                f"is given as input.\nCurrent input: \n{self.sd}"
            )
        else:
            self.sd = np.array(self.sd)
        if not np.issubdtype(self.sd.dtype, np.number):
            try:
                self.sd = self.sd.astype(float)
            except ValueError:
                raise ValueError(
                    f"The sd vector/matrix contains values that are not "
                    f"numeric. \nCurrent sd vector/matrix: \n{self.sd}"
                )
            except Exception as e:
                raise RuntimeError(f"Unknown error: {e}")
        else:
            # If the array is not the right shape, we assume it is a vector
            # that needs to be tiled into a matrix

            logger.debug(f"SD matrix: {self.sd}\n")
            logger.debug(f"Type of SD matrix: {type(self.sd)}")

            if self.sd.shape != self.experimental_matrix.shape:
                try:
                    self._build_sd_matrix()
                except ValueError:
                    raise
                except RuntimeError:
                    raise
            else:
                logger.debug(f"sd matrix: {self.sd}\n")
                return
        logger.info(f"sd Matrix:\n{self.sd}\n")

    def _build_sd_matrix(self):
        """
        Build the sd matrix from different input types

        :return: None
        """

        # First condition: the sds are in a 1D array
        if isinstance(self.sd, np.ndarray):
            # We first check that the sd vector is as long as the
            # experimental matrix on the row axis
            if self.sd.size != self.experimental_matrix[0].size:
                raise ValueError("sd vector not of right size")
            else:
                # We duplicate the vector column-wise to build a matrix of
                # duplicated sd vectors
                self.sd = np.tile(self.sd, (len(self.experimental_matrix), 1))

        # Second condition: the sd is a scalar and must be broadcast to a
        # matrix with same shape as the data
        elif isinstance(self.sd, int) or isinstance(self.sd, float):
            self.sd = np.full(self.experimental_matrix.shape, self.sd)
        else:
            raise RuntimeError("Unknown error")

    def _get_default_sds(self):
        """
        Build a default sd matrix. Default values:
            * Biomass: 0.2
            * Metabolites: 0.5
        :return: None
        """

        sds = [0.2]
        for name in range(len(self.model.name_vector) - 1):
            sds.append(0.5)
        self.sd = np.array(sds)
        self._build_sd_matrix()

    def optimize(self):
        """
        Run optimization and build the simulated matrix
        from the optimized parameters
        """

        logger.info("\nRunning optimization...\n")
        bounds = self.model.bounds()
        parameters = [param for param in self.model.parameters_to_estimate.values()]
        logger.debug(f"Simulate function = {self.model.simulate}")
        self.optimize_results = self._run_optimization(
            params=parameters,
            func=self.model.simulate,
            exp_data_matrix=self.experimental_matrix,
            time_vector=self.model.time_vector,
            non_opt_params=self.model.fixed_parameters,
            sd_matrix=self.sd,
            bounds=bounds,
            method="differential_evolution"
        )
        self.parameter_stats = {
            "optimal": self.optimize_results.x
        }
        logger.info(f"Optimization results: \n{self.optimize_results}\n")
        for i, param in zip(
                self.model.parameters_to_estimate, self.optimize_results.x
        ):
            logger.info(f"\n{i} = {param}\n")
        self.simulated_matrix = self.model.simulate(
            self.optimize_results.x,
            self.experimental_matrix,
            self.model.time_vector,
            self.model.fixed_parameters
        )
        logger.debug(f"simulated_matrix:\n{self.simulated_matrix}")
        nan_sim_mat = np.copy(self.simulated_matrix)
        nan_sim_mat[np.isnan(self.experimental_matrix)] = np.nan
        self.simulated_data = DataFrame(
            data=nan_sim_mat,
            index=self.model.time_vector,
            columns=self.model.name_vector
        )
        self.simulated_data.index.name = "Time"
        logger.info(f"Final Simulated Data: \n{self.simulated_data}\n")

    @staticmethod
    def _calculate_cost(
            params, func, exp_data_matrix, time_vector, non_opt_params,
            sd_matrix
    ):
        """
        Calculate the cost (residue) using the square of
        simulated-experimental over the SDs
        """

        simulated_matrix = func(params, exp_data_matrix, time_vector,
                                non_opt_params)
        cost_val = np.square((simulated_matrix - exp_data_matrix) / sd_matrix)
        residuum = np.nansum(cost_val)
        return residuum

    @staticmethod
    def _run_optimization(
            params: list,
            func: Model.simulate,
            exp_data_matrix: np.ndarray,
            time_vector: np.ndarray,
            non_opt_params: dict,
            sd_matrix: np.ndarray,
            bounds: tuple,
            method: str
    ):
        """
        Run the optimization on input parameters using the cost function and
        Scipy minimize (L-BFGS-B method that is deterministic and uses the
        gradient method for optimizing)
        """

        if method == "differential_evolution":
            logger.debug(f"Optimization method = {method}")
            optimize_results = differential_evolution(
                PhysioFitter._calculate_cost, bounds=bounds,
                args=(
                func, exp_data_matrix, time_vector, non_opt_params, sd_matrix),
                polish=True, x0=np.array(params)
            )
        elif method == "L-BFGS-B":
            logger.debug(f"Optimization method = {method}")
            optimize_results = minimize(
                PhysioFitter._calculate_cost, x0=np.array(params),
                args=(
                func, exp_data_matrix, time_vector, non_opt_params, sd_matrix),
                method="L-BFGS-B", bounds=bounds, options={'maxcor': 30}
            )
        else:
            raise ValueError(f"{method} is not implemented")

        return optimize_results

    def monte_carlo_analysis(self):
        """
        Run a monte carlo analysis to calculate optimization standard
        deviations on parameters and simulated data points
        """

        if not self.optimize_results:
            raise RuntimeError(
                "Running Monte Carlo simulation without having run the "
                "optimization is impossible as best fit results are needed to "
                "generate the initial simulated matrix"
            )

        logger.info(
            f"Running monte carlo analysis. Number of iterations: "
            f"{self.iterations}\n"
        )

        # Store the optimized results in variable that will be overridden on
        # every pass
        opt_res = self.optimize_results
        opt_params_list = []
        matrices = []

        for _ in range(self.iterations):
            new_matrix = self._apply_noise()

            # We optimise the parameters using the noisy matrix as input
            opt_res = PhysioFitter._run_optimization(
                opt_res.x, self.model.simulate, new_matrix, self.model.time_vector,
                self.model.fixed_parameters, self.sd, self.model.bounds(),
                "L-BFGS-B"
            )

            # Store the new simulated matrix in list for later use
            matrices.append(
                self.model.simulate(
                    opt_res.x, new_matrix, self.model.time_vector,
                    self.model.fixed_parameters
                )
            )

            # Store the new optimised parameters in list for later use
            opt_params_list.append(opt_res.x)

        # Build a 3D array from all the simulated matrices to get standard
        # deviation on each data point
        matrices = np.array(matrices)
        self.matrices_ci = {
            "lower_ci": np.percentile(matrices, 2.5, axis=0),
            "higher_ci": np.percentile(matrices, 97.5, axis=0)
        }

        # Compute the statistics on the list of parameters: means, sds,
        # medians and confidence interval
        self._compute_parameter_stats(opt_params_list)
        logger.info(f"Optimized parameters statistics:")
        for key, value in self.parameter_stats.items():
            logger.info(f"{key}: {value}")

        # Apply nan mask to be coherent with the experimental matrix
        nan_lower_ci = np.copy(self.matrices_ci['lower_ci'])
        nan_higher_ci = np.copy(self.matrices_ci['higher_ci'])
        nan_lower_ci[np.isnan(self.experimental_matrix)] = np.nan
        nan_higher_ci[np.isnan(self.experimental_matrix)] = np.nan

        logger.info(
            f"Simulated matrix lower confidence interval:\n{nan_lower_ci}\n"
        )
        logger.info(
            f"Simulated matrix higher confidence interval:\n{nan_higher_ci}\n"
        )
        return

    def _compute_parameter_stats(self, opt_params_list):
        """
        Compute statistics on the optimized parameters from the monte carlo
        analysis.

        :param opt_params_list: list of optimized parameter arrays generated
                                during the monte carlo analysis
        :return: parameter stats attribute containing means, sds, medians, low
                 and high CI
        """

        opt_params_means = np.mean(np.array(opt_params_list), 0)
        opt_params_sds = np.std(np.array(opt_params_list), 0)
        opt_params_meds = np.median(np.array(opt_params_list), 0)
        conf_ints = np.column_stack((
            np.percentile(opt_params_list, 2.5, 0),
            np.percentile(opt_params_list, 97.5, 0)
        ))

        self.parameter_stats.update({
            "mean": opt_params_means,
            "sd": opt_params_sds,
            "median": opt_params_meds,
            "CI_2.5": conf_ints[:, 0],
            "CI_97.5": conf_ints[:, 1]
        })

        # self.parameter_stats_df = DataFrame()

    def khi2_test(self):

        number_measurements = np.count_nonzero(
            ~np.isnan(self.experimental_matrix)
        )
        number_params = len(self.model.parameters_to_estimate)
        dof = number_measurements - number_params
        cost = self._calculate_cost(
            self.optimize_results.x, self.model.simulate, self.experimental_matrix,
            self.model.time_vector, self.model.fixed_parameters, self.sd
        )
        p_val = chi2.cdf(cost, dof)

        khi2_res = {
            "khi2_value": cost,
            "number_of_measurements": number_measurements,
            "number_of_params": number_params,
            "Degrees_of_freedom": dof,
            "p_val": p_val
        }
        self.khi2_res = DataFrame.from_dict(
            khi2_res, "index", columns=["Values"]
        )

        logger.info(f"khi2 test results:\n"
                     f"khi2 value: {cost}\n"
                     f"Number of measurements: {number_measurements}\n"
                     f"Number of parameters to fit: {number_params}\n"
                     f"Degrees of freedom: {dof}\n"
                     f"p-value = {p_val}\n"
                    )

        if p_val < 0.95:
            logger.info(
                f"At level of 95% confidence, the model fits the data good "
                f"enough with respect to the provided measurement SD. "
                f"Value: {p_val}"
            )

        else:
            logger.info(
                f"At level of 95% confidence, the model does not fit the data "
                f"good enough with respect to the provided measurement SD. "
                f"Value: {p_val}\n"
            )

    @staticmethod
    def _add_noise(vector, sd):
        """
        Add random noise to a given array using input standard deviations.

        :param vector: input array on which to apply noise
        :type vector: class: numpy.ndarray
        :param sd: standard deviation to apply to the input array
        :type sd: class: numpy.ndarray
        :return: noisy ndarray
        """

        output = np.random.default_rng().normal(
            loc=vector, scale=sd, size=vector.size
        )
        return output

    def _apply_noise(self):
        """
        Apply noise to the simulated matrix obtained using optimized
        parameters. SDs are obtained from the sd matrix
        """

        new_matrix = np.array([
            PhysioFitter._add_noise(self.simulated_matrix[idx, :], sd)
            for idx, sd in enumerate(self.sd)
        ])
        return new_matrix



"""
Module containing the methods used by PhysioFit.
"""

from __future__ import annotations
import numpy as np

from physiofit.models.base_model import Model, Bounds


class ChildModel(Model):

    def __init__(self, data):

        super().__init__(data)
        self.model_name = "Steady-state batch model with degradation of metabolites"
        self.vini = 1
        self.parameters_to_estimate = None
        self.fixed_parameters = None

    def get_params(self):

        self.parameters_to_estimate = {
            "X_O" : self.vini,
            "growth_rate" : self.vini
        }
        self.fixed_parameters = {"Degradation": {
            met: 0 for met in self.metabolites
            }
        }
        self.bounds = Bounds({
            "X_0": (1e-3, 10),
            "growth_rate": (1e-3, 3),
        })
        for metabolite in self.metabolites:
            self.parameters_to_estimate.update(
                {
                    f"{metabolite}_q": self.vini,
                    f"{metabolite}_M0": self.vini
                }
            )
            self.bounds.update(
                {
                    f"{metabolite}_q": (-50, 50),
                    f"{metabolite}_M0": (1e-6, 50)
                }
            )
            self.fixed_parameters = {"Degradation": {
                met: 0 for met in self.metabolites
                }
            }

    @staticmethod
    def simulate(
            params_opti: list,
            data_matrix: np.ndarray,
            time_vector: np.ndarray,
            params_non_opti: dict | list
    ):
        # Get end shape
        simulated_matrix = np.empty_like(data_matrix)

        # Get initial params
        x_0 = params_opti[0]
        mu = params_opti[1]

        # Get X_0 values
        exp_mu_t = np.exp(mu * time_vector)
        simulated_matrix[:, 0] = x_0 * exp_mu_t
        fixed_params = [value for value in params_non_opti["Degradation"].values()]

        for i in range(1, int(len(params_opti) / 2)):
            q = params_opti[i * 2]
            m_0 = params_opti[i * 2 + 1]
            k = fixed_params[i - 1]
            exp_k_t = np.exp(-k * time_vector)
            simulated_matrix[:, i] = q * (x_0 / (mu + k)) \
                                     * (exp_mu_t - exp_k_t) \
                                     + m_0 * exp_k_t

        return simulated_matrix



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['physiofit',
 'physiofit.base',
 'physiofit.models',
 'physiofit.tests',
 'physiofit.ui']

package_data = \
{'': ['*'], 'physiofit': ['data/*']}

install_requires = \
['matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=2.0.1,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'scipy>=1.10.1,<2.0.0',
 'streamlit>=1.20.0,<2.0.0']

entry_points = \
{'console_scripts': ['physiofit = physiofit.__main__:main']}

setup_kwargs = {
    'name': 'physiofit',
    'version': '3.3.1',
    'description': 'Calculate extracellular fluxes from metabolite concentrations and biomass data',
    'long_description': '# PhysioFit\n\n[![PyPI version](https://badge.fury.io/py/physiofit.svg)](https://badge.fury.io/py/physiofit)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/physiofit.svg)](https://pypi.python.org/pypi/physiofit/)\n[![Documentation Status](https://readthedocs.org/projects/physiofit/badge/?version=latest)](http://physiofit.readthedocs.io/?badge=latest)\n[![install with bioconda](https://img.shields.io/badge/install%20with-bioconda-brightgreen.svg?style=flat)](http://bioconda.github.io/recipes/physiofit/README.html)\n\n\n## What is PhysioFit?\n**PhysioFit is a scientific tool designed to quantify cell growth parameters and uptake & production fluxes**\n\nFluxes are estimated using mathematical models by fitting time-course measurements of the concentration of\ncells and extracellular substrates and products. PhysioFit is shipped with some common growth models, and\nadditional tailor-made models can be implemented by users.\n\nIt is one of the routine tools that we use at the [MetaSys team](https://www.toulouse-biotechnology-institute.fr/en/poles/equipe-metasys/) \nand [MetaToul platform](https://www.metabohub.fr/home.html) in functional studies of metabolic systems.\n\nThe code is open-source, and available under a GPLv3 license. Additional information can be found in the following \n[publication](https://doi.org/10.1101/2023.10.12.561695).\n\nDetailed documentation can be found online at Read the Docs \n([https://physiofit.readthedocs.io/](https://physiofit.readthedocs.io/)).\n\n## Key features\n\n   * **calculation of growth rate and extracellular (uptake and production) fluxes**,\n   * **a set of steady-state and dynamic models** are shipped with PhysioFit,\n   * **tailor-made models** can be constructed by users,\n   * Monte-Carlo sensitivity analysis to **estimate the precision on the calculated fluxes**,\n   * **evaluation of the goodness of fit and visual inspection of the fitted curves**,\n   * shipped as a **library** with both a **graphical** and a **command line** interface,\n   * **open-source, free and easy to install** everywhere where Python 3 and pip run,\n   * **biologist-friendly**.\n\n## Quick-start\nPhysioFit requires Python 3.9 or higher and run on all platforms.\nPlease check [the documentation](https://physiofit.readthedocs.io/en/latest/quickstart.html) for complete\ninstallation and usage instructions.\n\nUse `pip` to **install PhysioFit from PyPi**:\n\n```bash\n$ pip install physiofit\n```\n\nThen, start the graphical interface with:\n\n```bash\n$ physiofit\n```\n\nPhysioFit is also available directly from command-line and as a Python library.\n\n## Bug and feature requests\nIf you have an idea on how we could improve PhysioFit please submit a new *issue*\nto [our GitHub issue tracker](https://github.com/MetaSys-LISBP/PhysioFit/issues).\n\n\n## Developers guide\n### Contributions\nContributions are very welcome! :heart:\n\nPlease work on your own fork,\nfollow [PEP8](https://www.python.org/dev/peps/pep-0008/) style guide,\nand make sure you pass all the tests before a pull request.\n\n### Local install with pip\nIn development mode, do a `pip install -e /path/to/PhysioFit` to install\nlocally the development version.\n\n### Build the documentation locally\nBuild the HTML documentation with:\n\n```bash\n$ cd doc\n$ make html\n```\n\nThe PDF documentation can be built locally by replacing `html` by `latexpdf`\nin the command above. You will need a recent latex installation.\n\n## How to cite\nPhysioFit: a software to quantify cell growth parameters and extracellular fluxes.\nLe Grégam L., Guitton Y., Bellvert F., Jourdan F., Portais J.C., Millard P.\nbioRxiv preprint, [doi: 10.1101/2023.10.12.561695](https://doi.org/10.1101/2023.10.12.561695)\n\n## Authors\nLoïc Le Grégam, Pierre Millard\n\n## Contact\n:email: legregam@insa-toulouse.fr, millard@insa-toulouse.fr\n',
    'author': 'llegregam',
    'author_email': 'legregam@insa-toulouse.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8, !=2.7.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.7.*',
}


setup(**setup_kwargs)
